import numpy as np
import pytest
from test_inputs.catchment import *
from test_inputs.readers import *

import earthkit.hydro as ekh


@pytest.mark.parametrize(
    "river_network",
    [
        ("d8_ldd", d8_ldd_1),
        ("cama_downxy", cama_downxy_1),
        ("cama_nextxy", cama_nextxy_1),
        ("d8_ldd", d8_ldd_2),
        ("cama_downxy", cama_downxy_2),
        ("cama_nextxy", cama_nextxy_2),
    ],
    indirect=True,
)
def test_find_subcatchments_does_not_overwrite(river_network):
    field = np.arange(river_network.n_nodes) + 1
    subcatchments = ekh.find_subcatchments(river_network, field)
    print(subcatchments)
    print(field)
    np.testing.assert_array_equal(subcatchments, field)


@pytest.mark.parametrize(
    "river_network",
    [
        ("d8_ldd", d8_ldd_1),
        ("cama_downxy", cama_downxy_1),
        ("cama_nextxy", cama_nextxy_1),
        ("d8_ldd", d8_ldd_2),
        ("cama_downxy", cama_downxy_2),
        ("cama_nextxy", cama_nextxy_2),
    ],
    indirect=True,
)
def test_find_subcatchments_does_not_overwrite_2d(river_network):
    field = np.zeros(river_network.mask.shape, dtype="int")
    field[river_network.mask] = np.arange(river_network.n_nodes) + 1
    find_subcatchments = ekh.find_subcatchments(river_network, field)
    print(find_subcatchments)
    print(field)
    np.testing.assert_array_equal(find_subcatchments, field)


@pytest.mark.parametrize(
    "river_network, query_field, subcatchment",
    [
        (("d8_ldd", d8_ldd_1), catchment_query_field_1, subcatchment_1),
        (("cama_downxy", cama_downxy_1), catchment_query_field_1, subcatchment_1),
        (("cama_nextxy", cama_nextxy_1), catchment_query_field_1, subcatchment_1),
        (("d8_ldd", d8_ldd_2), catchment_query_field_2, subcatchment_2),
        (("cama_downxy", cama_downxy_2), catchment_query_field_2, subcatchment_2),
        (("cama_nextxy", cama_nextxy_2), catchment_query_field_2, subcatchment_2),
    ],
    indirect=["river_network"],
)
def test_find_subcatchments(river_network, query_field, subcatchment):
    subcatchments = ekh.find_subcatchments(river_network, query_field)
    print(subcatchment)
    print(subcatchments)
    np.testing.assert_array_equal(subcatchment, subcatchments)


@pytest.mark.parametrize(
    "river_network, query_field, find_subcatchments",
    [
        (("d8_ldd", d8_ldd_1), catchment_query_field_1, subcatchment_1),
        (("cama_downxy", cama_downxy_1), catchment_query_field_1, subcatchment_1),
        (("cama_nextxy", cama_nextxy_1), catchment_query_field_1, subcatchment_1),
        (("d8_ldd", d8_ldd_2), catchment_query_field_2, subcatchment_2),
        (("cama_downxy", cama_downxy_2), catchment_query_field_2, subcatchment_2),
        (("cama_nextxy", cama_nextxy_2), catchment_query_field_2, subcatchment_2),
    ],
    indirect=["river_network"],
)
def test_find_subcatchments_2d(river_network, query_field, find_subcatchments):
    field = np.zeros(river_network.mask.shape, dtype="int")
    field[river_network.mask] = query_field
    network_find_subcatchments = ekh.find_subcatchments(river_network, field)
    print(find_subcatchments)
    print(network_find_subcatchments)
    np.testing.assert_array_equal(
        network_find_subcatchments[river_network.mask], find_subcatchments
    )
    np.testing.assert_array_equal(network_find_subcatchments[~river_network.mask], 0)


@pytest.mark.parametrize(
    "river_network, query_field, find_subcatchments",
    [
        (("d8_ldd", d8_ldd_1), catchment_query_field_1, subcatchment_1),
        (("cama_downxy", cama_downxy_1), catchment_query_field_1, subcatchment_1),
        (("cama_nextxy", cama_nextxy_1), catchment_query_field_1, subcatchment_1),
        (("d8_ldd", d8_ldd_2), catchment_query_field_2, subcatchment_2),
        (("cama_downxy", cama_downxy_2), catchment_query_field_2, subcatchment_2),
        (("cama_nextxy", cama_nextxy_2), catchment_query_field_2, subcatchment_2),
    ],
    indirect=["river_network"],
)
def test_find_subcatchments_Nd(river_network, query_field, find_subcatchments):
    field = np.zeros(river_network.mask.shape, dtype="int")
    field[river_network.mask] = query_field
    field = np.stack([field, field], axis=0)
    network_find_subcatchments = ekh.find_subcatchments(river_network, field)
    find_subcatchments = np.stack([find_subcatchments, find_subcatchments], axis=0)
    print(find_subcatchments)
    print(network_find_subcatchments)
    np.testing.assert_array_equal(
        network_find_subcatchments[..., river_network.mask], find_subcatchments
    )
    np.testing.assert_array_equal(
        network_find_subcatchments[..., ~river_network.mask], 0
    )


@pytest.mark.parametrize(
    "river_network, query_field, find_catchments",
    [
        (("d8_ldd", d8_ldd_1), catchment_query_field_1, catchment_1),
        (("cama_downxy", cama_downxy_1), catchment_query_field_1, catchment_1),
        (("cama_nextxy", cama_nextxy_1), catchment_query_field_1, catchment_1),
        (("d8_ldd", d8_ldd_2), catchment_query_field_2, catchment_2),
        (("cama_downxy", cama_downxy_2), catchment_query_field_2, catchment_2),
        (("cama_nextxy", cama_nextxy_2), catchment_query_field_2, catchment_2),
    ],
    indirect=["river_network"],
)
def test_find_catchments_2d(river_network, query_field, find_catchments):
    field = np.zeros(river_network.mask.shape, dtype="int")
    field[river_network.mask] = query_field
    network_find_catchments = ekh.find_catchments(river_network, field)
    print(find_catchments)
    print(network_find_catchments)
    np.testing.assert_array_equal(
        network_find_catchments[river_network.mask], find_catchments
    )
    np.testing.assert_array_equal(network_find_catchments[~river_network.mask], 0)


@pytest.mark.parametrize(
    "river_network, query_field, find_catchments",
    [
        (("d8_ldd", d8_ldd_1), catchment_query_field_1, catchment_1),
        (("cama_downxy", cama_downxy_1), catchment_query_field_1, catchment_1),
        (("cama_nextxy", cama_nextxy_1), catchment_query_field_1, catchment_1),
        (("d8_ldd", d8_ldd_2), catchment_query_field_2, catchment_2),
        (("cama_downxy", cama_downxy_2), catchment_query_field_2, catchment_2),
        (("cama_nextxy", cama_nextxy_2), catchment_query_field_2, catchment_2),
    ],
    indirect=["river_network"],
)
def test_find_catchments_Nd(river_network, query_field, find_catchments):
    field = np.zeros(river_network.mask.shape, dtype="int")
    field[river_network.mask] = query_field
    field = np.stack([field, field], axis=0)
    network_find_catchments = ekh.find_catchments(river_network, field)
    find_catchments = np.stack([find_catchments, find_catchments], axis=0)
    print(find_catchments)
    print(network_find_catchments)
    np.testing.assert_array_equal(
        network_find_catchments[..., river_network.mask], find_catchments
    )
    np.testing.assert_array_equal(network_find_catchments[..., ~river_network.mask], 0)



import numpy as np


def extend_array(arr, extra_shape):
    current_shape = arr.shape
    new_shape = (*extra_shape, *current_shape)
    extended_array = np.broadcast_to(arr, new_shape)
    return extended_array


def convert_to_2d(river_network, array, fill_value):
    field = np.full(river_network.mask.shape, fill_value=fill_value, dtype=array.dtype)
    field[river_network.mask] = array
    return field



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np

from .utils import mask_and_unmask, missing_to_nan, nan_to_missing


@mask_and_unmask
def move_downstream(
    river_network,
    field,
    mv=np.nan,
    ufunc=np.add,
    accept_missing=False,
):
    """Sets each node to be the sum of its upstream nodes values, or a missing value.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field representing node values.
    mv : scalar, optional
        The missing value to use (default is np.nan).
    ufunc : callable, optional
        The numpy ufunc to perform when propagating (default is numpy.add).
    accept_missing : bool, optional
        If True, missing values are allowed in the input field (default is False).

    Returns
    -------
    numpy.ndarray
        The updated field with upstream contributions.

    """

    field, field_dtype = missing_to_nan(field, mv, accept_missing)

    ups = np.zeros(field.shape, dtype=field_dtype)
    mask = (
        river_network.downstream_nodes != river_network.n_nodes
    )  # remove sinks since they have no downstream
    nodes_to_update = river_network.downstream_nodes[mask]
    values_to_add = field[..., mask]
    ufunc.at(ups, (*[slice(None)] * (ups.ndim - 1), nodes_to_update), values_to_add)

    ups = nan_to_missing(ups, field_dtype, mv)

    return ups


@mask_and_unmask
def move_upstream(river_network, field, mv=np.nan, accept_missing=False):
    """Sets each node to be its downstream node value, or a missing value.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field representing node values.
    mv : scalar, optional
        The missing value to use (default is np.nan).
    accept_missing : bool, optional
        If True, missing values are allowed in the input field (default is False).

    Returns
    -------
    numpy.ndarray
        The updated field with downstream values.

    """

    field, field_dtype = missing_to_nan(field, mv, accept_missing)

    down = np.zeros(field.shape, dtype=field_dtype)
    mask = river_network.downstream_nodes != river_network.n_nodes  # remove sinks
    down[..., mask] = field[..., river_network.downstream_nodes[mask]]

    down = nan_to_missing(down, field_dtype, mv)

    return down



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from io import BytesIO
from urllib.request import urlopen

import joblib
import numpy as np

from ._version import __version__ as ekh_version
from .readers import (
    cache,
    find_main_var,
    from_cama_nextxy,
    from_d8,
    import_earthkit_or_prompt_install,
)

# read in only up to second decimal point
# i.e. 0.1.dev90+gfdf4e33.d20250107 -> 0.1
ekh_version = ".".join(ekh_version.split(".")[:2])


@cache
def create(path, river_network_format, source):
    """Creates a river network from the given path, format, and source.

    Parameters
    ----------
    path : str
        The path to the river network data.
    river_network_format : str
        The format of the river network data.
        Supported formats are "precomputed", "cama", "pcr_d8", "esri_d8"
        and "merit_d8".
    source : str
        The source of the river network data.
        For possible sources see:
        https://earthkit-data.readthedocs.io/en/latest/guide/sources.html.

    Returns
    -------
    earthkit.hydro.RiverNetwork
        The river network object created from the given data.

    """
    if river_network_format == "precomputed":
        if source == "file":
            return joblib.load(path)
        elif source == "url":
            return joblib.load(BytesIO(urlopen(path).read()))
        else:
            raise ValueError(
                "Unsupported source for river network format"
                f"{river_network_format}: {source}."
            )
    elif river_network_format == "cama":
        ekd = import_earthkit_or_prompt_install(river_network_format, source)
        data = ekd.from_source(source, path).to_xarray(mask_and_scale=False)
        x, y = data.nextx.values, data.nexty.values
        return from_cama_nextxy(x, y)
    elif (
        river_network_format == "pcr_d8"
        or river_network_format == "esri_d8"
        or river_network_format == "merit_d8"
    ):
        if path.endswith(".map"):
            from .pcr import from_file

            data = from_file(path, mask=False)
        else:
            ekd = import_earthkit_or_prompt_install(river_network_format, source)
            data = ekd.from_source(source, path).to_xarray(mask_and_scale=False)
            var_name = find_main_var(data)
            data = data[var_name].values
        return from_d8(data, river_network_format=river_network_format)
    else:
        raise ValueError(f"Unsupported river network format: {river_network_format}.")


def load(
    domain,
    river_network_version,
    data_source=(
        "https://github.com/ecmwf/earthkit-hydro-store/raw/refs/heads/main/"
        "{ekh_version}/{domain}/{river_network_version}/river_network.joblib"
    ),
    *args,
    **kwargs,
):
    """Load a precomputed river network from a named domain and
    river_network_version.

    Parameters
    ----------
    domain : str
        The domain of the river network. Supported domains are "efas", "glofas",
        "cama_15min", "cama_06min", "cama_05min", "cama_03min".
    river_network_version : str
        The version of the river network on the specified domain.
    data_source : str, optional
        The data source URL template for the river network.
    *args : tuple
        Additional positional arguments to pass to `create_river_network`.
    **kwargs : dict
        Additional keyword arguments to pass to `create_river_network`.

    Returns
    -------
    earthkit.hydro.RiverNetwork
        The loaded river network.

    """
    uri = data_source.format(
        ekh_version=ekh_version,
        domain=domain,
        river_network_version=river_network_version,
    )
    network = create(uri, "precomputed", "url", *args, **kwargs)
    if network.sources.dtype != np.uintp:
        network.sources = network.sources.astype(np.uintp)
    if network.downstream_nodes.dtype != np.uintp:
        network.downstream_nodes = network.downstream_nodes.astype(np.uintp)
    if network.sinks.dtype != np.uintp:
        network.sinks = network.sinks.astype(np.uintp)
    return network


def available():
    """
    Prints the available precomputed networks.
    """

    print(
        "Available precomputed networks are:\n",
        '`ekh.river_network.load("efas", "5")`\n',
        '`ekh.river_network.load("efas", "4")`\n',
        '`ekh.river_network.load("glofas", "4")`\n',
        '`ekh.river_network.load("glofas", "3")`\n',
        '`ekh.river_network.load("cama_15min", "4")`\n',
        '`ekh.river_network.load("cama_06min", "4")`\n',
        '`ekh.river_network.load("cama_05min", "4")`\n',
        '`ekh.river_network.load("cama_03min", "4")`\n',
        '`ekh.river_network.load("hydrosheds_06min", "1")`\n',
        '`ekh.river_network.load("hydrosheds_05min", "1")`',
    )



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np

from .core import flow
from .metrics import metrics_dict
from .utils import mask_and_unmask, missing_to_nan, nan_to_missing


@mask_and_unmask
def flow_downstream(
    river_network,
    field,
    mv=np.nan,
    in_place=False,
    ufunc=np.add,
    accept_missing=False,
    skip_missing_check=False,
    additive_weight=None,
    multiplicative_weight=None,
    modifier_use_upstream=True,
):
    """Accumulates field values downstream.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    mv : scalar, optional
        The missing value indicator. Default is np.nan.
    in_place : bool, optional
        If True, modifies the input field in place. Default is False.
    ufunc : numpy.ufunc, optional
        The universal function (ufunc) to use for accumulation. Default is np.add.
    accept_missing : bool, optional
        If True, accepts missing values in the field. Default is False.
    skip_missing_check : bool, optional
        Whether to skip checking for missing values. Default is False.
    additive_weight : numpy.ndarray, optional
        A weight to be added to the field values. Default is None.
    multiplicative_weight : numpy.ndarray, optional
        A weight to be multiplied with the field values. Default is None.
    modifier_use_upstream : bool, optional
        If True, the modifiers are used on the upstream nodes instead of downstream.
        Default is True.

    Returns
    -------
    numpy.ndarray
        The field values accumulated downstream.

    """

    if not in_place:
        field = field.copy()

    # TODO: decide if we should check all of the weights also
    field, field_dtype = missing_to_nan(field, mv, accept_missing, skip_missing_check)

    op = _ufunc_to_downstream

    def operation(
        river_network,
        field,
        grouping,
        mv,
        additive_weight,
        multiplicative_weight,
        modifier_use_upstream,
    ):
        return op(
            river_network,
            field,
            grouping,
            mv,
            additive_weight,
            multiplicative_weight,
            modifier_use_upstream,
            ufunc=ufunc,
        )

    field = flow(
        river_network,
        field,
        False,
        operation,
        mv,
        additive_weight,
        multiplicative_weight,
        modifier_use_upstream,
    )

    return nan_to_missing(field, field_dtype, mv)


def _ufunc_to_downstream(
    river_network,
    field,
    grouping,
    mv,
    additive_weight,
    multiplicative_weight,
    modifier_use_upstream,
    ufunc,
):
    """Updates field in-place by applying a ufunc at the downstream nodes of
    the grouping.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    grouping : numpy.ndarray
        An array of indices.
    mv : scalar
        A missing value indicator (not used in the function but kept for consistency).
    additive_weight : numpy.ndarray, optional
        A weight to be added to the field values. Default is None.
    multiplicative_weight : numpy.ndarray, optional
        A weight to be multiplied with the field values. Default is None.
    modifier_use_upstream : bool, optional
        If True, the modifiers are used on the upstream nodes instead of downstream.
        Default is True.
    ufunc : numpy.ufunc
        A universal function from the numpy library to be applied to the field data.
        Available ufuncs: https://numpy.org/doc/2.2/reference/ufuncs.html.
        Note: must allow two operands.

    Returns
    -------
    None

    """
    modifier_group = (
        grouping if modifier_use_upstream else river_network.downstream_nodes[grouping]
    )

    if additive_weight is None:
        if multiplicative_weight is None:
            modifier_field = field[..., grouping]
        else:
            modifier_field = field[grouping] * multiplicative_weight[modifier_group]
    else:
        if multiplicative_weight is None:
            modifier_field = field[grouping] + additive_weight[modifier_group]
        else:
            modifier_field = (
                field[grouping] * multiplicative_weight[modifier_group]
                + additive_weight[modifier_group]
            )

    ufunc.at(
        field,
        (*[slice(None)] * (field.ndim - 1), river_network.downstream_nodes[grouping]),
        modifier_field,
    )


@mask_and_unmask
def flow_upstream(
    river_network,
    field,
    mv=np.nan,
    in_place=False,
    ufunc=np.add,
    accept_missing=False,
    skip_missing_check=False,
    additive_weight=None,
    multiplicative_weight=None,
    modifier_use_upstream=True,
):
    """Accumulates field values upstream.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    mv : scalar, optional
        The missing value indicator. Default is np.nan.
    in_place : bool, optional
        If True, modifies the input field in place. Default is False.
    ufunc : numpy.ufunc, optional
        The universal function (ufunc) to use for accumulation. Default is np.add.
    accept_missing : bool, optional
        If True, accepts missing values in the field. Default is False.
    skip_missing_check : bool, optional
        Whether to skip checking for missing values. Default is False.
    additive_weight : numpy.ndarray, optional
        A weight to be added to the field values. Default is None.
    multiplicative_weight : numpy.ndarray, optional
        A weight to be multiplied with the field values. Default is None.
    modifier_use_upstream : bool, optional
        If True, the modifiers are used on the upstream nodes instead of downstream.
        Default is True.

    Returns
    -------
    numpy.ndarray
        The field values accumulated downstream.

    """

    if not in_place:
        field = field.copy()

    field, field_dtype = missing_to_nan(field, mv, accept_missing, skip_missing_check)

    op = _ufunc_to_upstream

    def operation(
        river_network,
        field,
        grouping,
        mv,
        additive_weight,
        multiplicative_weight,
        modifier_use_upstream,
    ):
        return op(
            river_network,
            field,
            grouping,
            mv,
            additive_weight,
            multiplicative_weight,
            modifier_use_upstream,
            ufunc=ufunc,
        )

    field = flow(
        river_network,
        field,
        True,
        operation,
        mv,
        additive_weight,
        multiplicative_weight,
        modifier_use_upstream,
    )

    return nan_to_missing(field, field_dtype, mv)


def _ufunc_to_upstream(
    river_network,
    field,
    grouping,
    mv,
    additive_weight,
    multiplicative_weight,
    modifier_use_upstream,
    ufunc,
):
    """Updates field in-place by applying a ufunc at the nodes of
    the grouping.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    grouping : numpy.ndarray
        An array of indices.
    mv : scalar
        A missing value indicator (not used in the function but kept for consistency).
    additive_weight : numpy.ndarray, optional
        A weight to be added to the field values. Default is None.
    multiplicative_weight : numpy.ndarray, optional
        A weight to be multiplied with the field values. Default is None.
    modifier_use_upstream : bool, optional
        If True, the modifiers are used on the upstream nodes instead of downstream.
        Default is True.
    ufunc : numpy.ufunc
        A universal function from the numpy library to be applied to the field data.
        Available ufuncs: https://numpy.org/doc/2.2/reference/ufuncs.html.
        Note: must allow two operands.

    Returns
    -------
    None

    """
    down_group = river_network.downstream_nodes[grouping]
    modifier_group = (
        grouping if modifier_use_upstream else river_network.downstream_nodes[grouping]
    )
    if additive_weight is None:
        if multiplicative_weight is None:
            modifier_field = field[..., down_group]
        else:
            modifier_field = (
                field[..., down_group] * multiplicative_weight[..., modifier_group]
            )
    else:
        if multiplicative_weight is None:
            modifier_field = (
                field[..., down_group] + additive_weight[..., modifier_group]
            )
        else:
            modifier_field = (
                field[..., down_group] * multiplicative_weight[..., modifier_group]
                + additive_weight[..., modifier_group]
            )

    ufunc.at(
        field,
        (*[slice(None)] * (field.ndim - 1), grouping),
        modifier_field,
    )


def calculate_online_metric(
    river_network, field, metric, weights, mv, accept_missing, flow_direction
):
    """
    Calculates a metric for the field over all upstream or downstream values.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    metric : str
        Metric to compute. Options are "mean", "max", "min", "sum", "product"
    weights : ndarray
        Used to weight the field when computing the metric.
    mv : scalar
        Missing value for the input field.
    accept_missing : bool
        Whether or not to accept missing values in the input field.
    flow_direction : str
        The direction of flow for the metric calculation. Options are "up" or "down".

    Returns
    -------
    numpy.ndarray
        Output field.

    """

    if flow_direction == "up":
        flow_func = flow_upstream
    elif flow_direction == "down":
        flow_func = flow_downstream

    field, field_dtype = missing_to_nan(field.copy(), mv, accept_missing)

    if weights is None:
        if metric == "mean" or metric == "std" or metric == "var":
            weightings = np.ones(river_network.n_nodes, dtype=np.float64)
        weighted_field = field.copy()
    else:
        assert field_dtype == weights.dtype
        weightings, _ = missing_to_nan(weights.copy(), mv, accept_missing)
        weighted_field = field * weightings  # this isn't in_place !

    ufunc = metrics_dict[metric].func

    weighted_field = flow_func(
        river_network,
        weighted_field,
        np.nan,  # mv replaced by nan
        True,  # do in-place on field copy
        ufunc,
        accept_missing,
        skip_missing_check=True,
        skip=True,
    )

    if metric == "mean" or metric == "std" or metric == "var":
        counts = flow_func(
            river_network,
            weightings,
            np.nan,  # mv replaced by nan
            False,
            ufunc,
            accept_missing,
            skip_missing_check=True,
            skip=True,
        )

        if metric == "mean":
            weighted_field /= counts  # weighted mean
            return nan_to_missing(
                weighted_field, np.float64, mv
            )  # if we compute means, we change dtype for int fields etc.
        elif metric == "var" or metric == "std":
            weighted_sum_of_squares = flow_func(
                river_network,
                field**2 * weightings if weights is not None else field**2,
                np.nan,  # mv replaced by nan
                True,  # do in-place on field copy
                ufunc,
                accept_missing,
                skip_missing_check=True,
                skip=True,
            )
            mean = weighted_field / counts
            weighted_sum_of_squares = weighted_sum_of_squares / counts - mean**2
            weighted_sum_of_squares[weighted_sum_of_squares < 0] = (
                0  # can occur for numerical issues
            )
            if metric == "var":
                return nan_to_missing(weighted_sum_of_squares, np.float64, mv)
            elif metric == "std":
                return nan_to_missing(np.sqrt(weighted_sum_of_squares), np.float64, mv)

    else:
        return nan_to_missing(weighted_field, field_dtype, mv)



import pytest

import earthkit.hydro as ekh


@pytest.fixture
def river_network(request):
    river_network_format, flow_directions = request.param
    if river_network_format == "d8_ldd":
        river_network = ekh.readers.from_d8(flow_directions)
    elif river_network_format == "cama_downxy":
        river_network = ekh.readers.from_cama_downxy(*flow_directions)
    elif river_network_format == "cama_nextxy":
        river_network = ekh.readers.from_cama_nextxy(*flow_directions)
    # TODO: add ESRI

    return river_network



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import earthkit.hydro.catchments  # for dynamic function creation
import earthkit.hydro.distance
import earthkit.hydro.length
import earthkit.hydro.pcr
import earthkit.hydro.readers  # for tests
import earthkit.hydro.river_network
import earthkit.hydro.subcatchments  # for dynamic function creation
import earthkit.hydro.upstream  # for dynamic function creation
import earthkit.hydro.zonal  # for dynamic function creation
from earthkit.hydro.accumulation import flow_downstream, flow_upstream
from earthkit.hydro.catchments import calculate_catchment_metric
from earthkit.hydro.catchments import find as find_catchments
from earthkit.hydro.downstream import calculate_downstream_metric
from earthkit.hydro.movement import move_downstream, move_upstream
from earthkit.hydro.river_network import create as create_river_network
from earthkit.hydro.river_network import load as load_river_network
from earthkit.hydro.subcatchments import calculate_subcatchment_metric
from earthkit.hydro.subcatchments import find as find_subcatchments
from earthkit.hydro.upstream import calculate_upstream_metric
from earthkit.hydro.zonal import calculate_zonal_metric

from ._version import __version__



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from struct import unpack

import numpy as np

# CSF value scales
# version 2 datatypes
VS_BOOLEAN = 0xE0  # boolean, always UINT1, values: 0,1 or MV_UINT1
VS_NOMINAL = 0xE2  # nominal, UINT1 or INT4
VS_ORDINAL = 0xF2  # ordinal, UINT1 or INT4
VS_SCALAR = 0xEB  # scalar, REAL4 or (maybe) REAL8
VS_DIRECTION = 0xFB  # directional REAL4 or (maybe) REAL8, -1 means no direction
VS_LDD = 0xF0  # local drain direction, always UINT1, values: 1-9 or MV_UINT1
# this one CANNOT be returned by NOR passed to a csf2 function
VS_UNDEFINED = 100  # just some value different from the rest

# CSF cell representations
# preferred version 2 cell representations
CR_UINT1 = 0x00  # boolean, ldd and small nominal and small ordinal
CR_INT4 = 0x26  # large nominal and large ordinal
CR_REAL4 = 0x5A  # single scalar and single directional
# other version 2 cell representations
CR_REAL8 = 0xDB  # double scalar or directional, no loss of precision


def _replace_missing_u1(cur, new):
    out = np.copy(cur)
    out[cur == 255] = new
    return out


def _replace_missing_i4(cur, new):
    out = np.copy(cur)
    out[cur == -2147483648] = new
    return out


def _replace_missing_f4(cur, new):
    out = np.copy(cur)
    out[np.isnan(cur)] = new
    return out


def _replace_missing_f8(cur, new):
    out = np.copy(cur)
    out[np.isnan(cur)] = new
    return out


CELLREPR = {
    CR_UINT1: dict(
        dtype=np.dtype("uint8"),
        fillmv=_replace_missing_u1,
    ),
    CR_INT4: dict(
        dtype=np.dtype("int32"),
        fillmv=_replace_missing_i4,
    ),
    CR_REAL4: dict(
        dtype=np.dtype("float32"),
        fillmv=_replace_missing_f4,
    ),
    CR_REAL8: dict(
        dtype=np.dtype("float64"),
        fillmv=_replace_missing_f8,
    ),
}


def from_file(path, mask=False):
    """
    Load a .map file into a numpy array.
    """

    with open(path, "rb") as f:
        bytes = f.read()

    nbytes_header = 64 + 2 + 2 + 8 + 8 + 8 + 8 + 4 + 4 + 8 + 8 + 8
    _, cellRepr, _, _, _, _, nrRows, nrCols, _, _, _ = unpack(
        "=hhddddIIddd", bytes[64:nbytes_header]
    )

    try:
        celltype = CELLREPR[cellRepr]
    except KeyError:
        raise Exception(
            "{}: invalid cellRepr value ({}) in header".format(path, cellRepr)
        )

    dtype = celltype["dtype"]

    size = dtype.itemsize * nrRows * nrCols
    data = np.frombuffer(bytes[256 : 256 + size], dtype)
    if mask:
        data = celltype["fillmv"](data.astype(np.float64), np.nan)

    data = data.reshape((nrRows, nrCols))

    return data



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import joblib
import numpy as np

from .utils import mask_2d


class RiverNetwork:
    """A class representing a river network for hydrological processing.

    Attributes
    ----------
    nodes : numpy.ndarray
        Array containing the node ids of the river network.
    n_nodes : int
        The number of nodes in the river network.
    downstream_nodes : numpy.ndarray
        Array of downstream node ids corresponding to each node.
    mask : numpy.ndarray
        A boolean mask converting from the domain to the river network.
    sinks : numpy.ndarray
        Nodes with no downstream connections.
    sources : numpy.ndarray
        Nodes with no upstream connections.
    topological_labels : numpy.ndarray
        A topological group label for each node.
    topological_groups : list of numpy.ndarray
        Groups of nodes sorted in topological order.

    """

    def __init__(
        self,
        nodes,
        downstream,
        mask,
        sinks=None,
        sources=None,
        topological_labels=None,
        check_for_cycles=False,
    ) -> None:
        """Initialises the RiverNetwork with nodes, downstream nodes, and a
        mask.

        Parameters
        ----------
        nodes : numpy.ndarray
            Array containing the node ids of the river network.
        downstream : numpy.ndarray
            Array of downstream node ids corresponding to each node.
        mask : numpy.ndarray
            A mask converting from the domain to the river network.
        sinks : numpy.ndarray, optional
            Array of sinks of the river network.
        sources :  numpy.ndarray, optional
            Array of sources of the river network.
        topological_labels : numpy.ndarray, optional
            Array of precomputed topological distance labels.
        check_for_cycles : bool, optional
            Whether to check for cycles when instantiating the river network.

        """
        self.nodes = nodes
        del nodes
        self.n_nodes = len(self.nodes)
        self.downstream_nodes = downstream
        del downstream
        self.mask = mask
        del mask
        self.sinks = (
            sinks
            if sinks is not None
            else self.nodes[self.downstream_nodes == self.n_nodes]
        )  # nodes with no downstreams
        del sinks
        self.sources = (
            sources if sources is not None else self.get_sources()
        )  # nodes with no upstreams
        del sources
        if check_for_cycles:
            self.check_for_cycles()
        self.topological_labels = (
            topological_labels
            if topological_labels is not None
            else self.compute_topological_labels()
        )
        del topological_labels
        self.topological_groups = self.topological_groups_from_labels()

    @property
    def shape(self):
        return self.mask.shape

    def get_sources(self):
        """Identifies the source nodes in the river network (nodes with no
        upstream nodes).

        Returns
        -------
        numpy.ndarray
            Array of source nodes.

        """
        tmp_nodes = self.nodes.copy()
        downstream_no_sinks = self.downstream_nodes[
            self.downstream_nodes != self.n_nodes
        ]  # get all downstream nodes
        tmp_nodes[downstream_no_sinks] = (
            self.n_nodes + 1
        )  # downstream nodes that aren't sinks = -1
        inlets = tmp_nodes[
            tmp_nodes != self.n_nodes + 1
        ]  # sources are nodes that are not downstream nodes
        return inlets

    def check_for_cycles(self):
        """Checks if the river network contains any cycles and raises an
        Exception if it does.
        """
        nodes = self.downstream_nodes.copy()
        while True:
            if np.any(nodes == self.nodes):
                Exception("River Network contains a cycle.")
            elif np.all(nodes == self.n_nodes):
                break
            not_sinks = nodes != self.n_nodes
            nodes[not_sinks] = self.downstream_nodes[nodes[not_sinks]].copy()
        print("No cycles found in the river network.")

    def compute_topological_labels(self):
        """Finds the topological distance labels for each node in the river
        network.

        Returns
        -------
        numpy.ndarray
            Array of topological distance labels for each node.

        """
        try:
            from .topological_labels_rust import compute_topological_labels
        except (ModuleNotFoundError, ImportError):
            print(
                "Failed to load rust extension, falling back to python implementation."
            )
            from .topological_labels_python import compute_topological_labels
        return compute_topological_labels(
            self.sources, self.sinks, self.downstream_nodes
        )

    def topological_groups_from_labels(self):
        """Groups nodes by their topological distance labels.

        Parameters
        ----------
        labels : numpy.ndarray
            Array of labels representing the topological distances of nodes.

        Returns
        -------
        list of numpy.ndarray
            A list of subarrays, each containing nodes with the same label.

        """
        sorted_indices = np.argsort(self.topological_labels)  # sort by labels
        sorted_array = self.nodes[sorted_indices]
        sorted_labels = self.topological_labels[sorted_indices]
        _, indices = np.unique(
            sorted_labels, return_index=True
        )  # find index of first occurrence of each label
        subarrays = np.split(
            sorted_array, indices[1:]
        )  # split array at each first occurrence of a label
        return subarrays

    def export(self, fpath="river_network.joblib", compression=1):
        """Exports the river network instance to a file.

        Parameters
        ----------
        fpath : str, optional
            The filepath to save the instance (default is "river_network.joblib").
        compression : int, optional
            Compression level for joblib (default is 1).

        """
        joblib.dump(self, fpath, compress=compression)

    @mask_2d
    def create_subnetwork(self, field, recompute=False):
        """Creates a subnetwork from the river network based on a mask.

        Parameters
        ----------
        field : numpy.ndarray
            A boolean mask to subset the river network.
        recompute : bool, optional
            If True, recomputes the topological labels for the subnetwork.
            Default is False.

        Returns
        -------
        RiverNetwork
            A subnetwork of the river network.

        """
        river_network_mask = field
        valid_indices = np.where(self.mask)
        new_valid_indices = (
            valid_indices[0][river_network_mask],
            valid_indices[1][river_network_mask],
        )
        domain_mask = np.full(self.mask.shape, False)
        domain_mask[new_valid_indices] = True

        downstream_indices = self.downstream_nodes[river_network_mask]
        n_nodes = len(downstream_indices)  # number of nodes in the subnetwork
        # create new array of network nodes, setting all nodes not in mask to n_nodes
        subnetwork_nodes = np.full(self.n_nodes, n_nodes)
        subnetwork_nodes[river_network_mask] = np.arange(n_nodes)
        # get downstream nodes in the subnetwork
        non_sinks = np.where(downstream_indices != self.n_nodes)
        downstream = np.full(n_nodes, n_nodes, dtype=np.uintp)
        downstream[non_sinks] = subnetwork_nodes[downstream_indices[non_sinks]]
        nodes = np.arange(n_nodes, dtype=np.uintp)

        if not recompute:
            sinks = nodes[downstream == n_nodes]
            topological_labels = self.topological_labels[river_network_mask]
            topological_labels[sinks] = self.n_nodes

            return RiverNetwork(
                nodes,
                downstream,
                domain_mask,
                sinks=sinks,
                topological_labels=topological_labels,
            )
        else:
            return RiverNetwork(nodes, downstream, domain_mask)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from functools import partial

import numpy as np

from .metrics import metrics_dict
from .utils import is_missing, missing_to_nan, nan_to_missing


def calculate_zonal_metric(
    field,
    labels,
    metric,
    weights=None,
    field_mv=np.nan,
    labels_mv=0,
    return_field=False,
    field_accept_missing=False,
    skip_missing_check=False,
):
    """
    Calculate a metric for each label.

    Parameters
    ----------
    field : ndarray
        The field data.
    labels : ndarray
        The labels for the field.
    metric : str
        Metric to compute. Options are "mean", "max", "min", "sum", "product"
    weights : ndarray, optional
        Used to weight the field when computing the metric. Default is None.
    field_mv : scalar, optional
        The missing value for the input fields. Default is np.nan.
    labels_mv : scalar, optional
        The missing values for the labels. Default is 0.
    return_field : bool, optional
        If True, return as a full field. If False, return as dict. Default is False.
    field_accept_missing : bool, optional
        Whether or not to accept missing values in the input fields. Default is False.
    skip_missing_check : bool, optional
        Whether or not to skip checking for missing values. Default is False.

    Returns
    -------
    np.array or dict
        Field if return_field, else dictionary with (label, metric) pairs.
    """
    ufunc = metrics_dict[metric].func

    mask = ~is_missing(labels, labels_mv)
    not_missing_labels = labels[..., mask].T

    relevant_field = field[..., mask].T

    relevant_field, field_dtype = missing_to_nan(
        relevant_field, field_mv, field_accept_missing, skip_missing_check
    )

    if weights is not None:
        assert field_dtype == weights.dtype
        relevant_weights = weights[..., mask].T
        relevant_weights, _ = missing_to_nan(
            relevant_weights, field_mv, field_accept_missing, skip_missing_check
        )
        weighted_field = (relevant_field.T * relevant_weights.T).T
    else:
        weighted_field = relevant_field

    unique_labels, unique_label_positions = np.unique(
        not_missing_labels, return_inverse=True
    )

    initial_field = np.full(
        (len(unique_labels), *field.T.shape[labels.ndim :]),
        metrics_dict[metric].base_val,
        dtype=np.float64,
    )

    ufunc.at(
        initial_field,
        (unique_label_positions, *[slice(None)] * (initial_field.ndim - 1)),
        weighted_field,
    )

    if metric == "mean" or metric == "var" or metric == "std":
        if weights is None:
            count_values = np.bincount(
                unique_label_positions, minlength=len(unique_labels)
            ).astype(initial_field.dtype)
        else:
            count_values = np.full(
                (len(unique_labels), *weights.T.shape[labels.ndim :]),
                metrics_dict[metric].base_val,
                dtype=relevant_weights.dtype,
            )
            ufunc.at(
                count_values,
                (unique_label_positions, *[slice(None)] * (count_values.ndim - 1)),
                relevant_weights,
            )
        initial_field_T = initial_field.T
        initial_field_T /= count_values.T
        initial_field = initial_field_T.T

        field_dtype = np.float64

        if metric == "var" or metric == "std":
            out_field = np.full(
                (len(unique_labels), *field.T.shape[labels.ndim :]),
                metrics_dict[metric].base_val,
                dtype=np.float64,
            )
            ufunc.at(
                out_field,
                (unique_label_positions, *[slice(None)] * (out_field.ndim - 1)),
                (
                    (initial_field[unique_label_positions] - relevant_field) ** 2
                    if weights is None
                    else relevant_weights
                    * (initial_field[unique_label_positions] - relevant_field) ** 2
                ),
            )
            initial_field = (out_field.T / count_values.T).T
            if metric == "std":
                initial_field = np.sqrt(initial_field)

    initial_field = nan_to_missing(initial_field, field_dtype, field_mv)

    if np.result_type(field_mv, initial_field) != initial_field.dtype:
        raise ValueError(
            f"Missing value of type {type(field_mv)} is not compatible"
            f" with field of dtype {initial_field.dtype}"
        )

    if return_field:
        out_field = np.empty(field.shape, dtype=np.float64)
        out_field[..., ~mask] = np.nan  # works correctly
        out_field[..., mask] = initial_field[unique_label_positions].T
        return out_field
    else:
        initial_field = np.transpose(
            initial_field, axes=[0] + list(range(initial_field.ndim - 1, 0, -1))
        )
        return dict(zip(unique_labels, initial_field))


for metric in metrics_dict.keys():

    func = partial(calculate_zonal_metric, metric=metric)

    globals()[metric] = func



import pytest
from test_inputs.readers import *
from test_inputs.subnetwork import *

import earthkit.hydro as ekh


@pytest.mark.parametrize(
    "river_network, mask, accumulate_downstream",
    [
        (("d8_ldd", d8_ldd_2), mask_2, masked_unit_accuflux_2),
        (("cama_downxy", cama_downxy_2), mask_2, masked_unit_accuflux_2),
        (("cama_nextxy", cama_nextxy_2), mask_2, masked_unit_accuflux_2),
    ],
    indirect=["river_network"],
)
def test_subnetwork(river_network, mask, accumulate_downstream):
    river_network = river_network.create_subnetwork(mask)
    field = np.ones(river_network.n_nodes)
    accum = ekh.flow_downstream(river_network, field)
    print(accum)
    print(accumulate_downstream)
    np.testing.assert_array_equal(accum, accumulate_downstream)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np

from ._rust import propagate_labels


def compute_topological_labels(
    sources: np.ndarray, sinks: np.ndarray, downstream_nodes: np.ndarray
):
    """Finds the topological distance labels for each node in
    downstream_nodes.

    Parameters
    ----------
    sources : numpy.ndarray
        The river network sources.
    sinks : numpy.ndarray
        The river network sinks.
    downstream_nodes : cnumpy.ndarray
        The river network downstream nodes.

    Returns
    -------
    numpy.ndarray
        Array of topological distance labels for each node.
    """

    n_nodes = np.uintp(downstream_nodes.shape[0])
    labels = np.zeros(n_nodes, dtype=np.int64)

    labels = propagate_labels(labels, sources, sinks, downstream_nodes, n_nodes)

    return labels



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from functools import partial

import numpy as np

from .core import flow
from .metrics import metrics_dict
from .upstream import calculate_upstream_metric
from .utils import (
    is_missing,
    mask_2d,
    mask_and_unmask,
    points_to_1d_indices,
    points_to_numpy,
)


@mask_2d
def calculate_catchment_metric(
    river_network,
    field,
    points,
    metric,
    weights=None,
    mv=np.nan,
    accept_missing=False,
):
    """
    Calculates the metric over the catchments defined by the points.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    points : list of tuples
        List of tuple indices of the points.
    metric : str
        Metric to compute. Options are "mean", "max", "min", "sum", "prod",
        "std", "var".
    weights : ndarray, optional
        Used to weight the field when computing the metric. Default is None.
    mv : scalar, optional
        Missing value for the input field. Default is np.nan.
    accept_missing : bool, optional
        Whether or not to accept missing values in the input field. Default is False.

    Returns
    -------
    dict
        Dictionary with (station, catchment_metric) pairs.

    """
    # TODO: Future idea could be to find all
    # nodes relevant for computing upstream
    # average, then creating a river subnetwork
    # and calculating upstream metric only there
    # (should be quicker, particularly for
    # small numbers of points)

    if isinstance(points, np.ndarray):
        upstream_metric_field = calculate_upstream_metric(
            river_network,
            field,
            metric,
            weights,
            mv,
            accept_missing,
            skip=True,
        )
        upstream_field_at_stations = upstream_metric_field[..., points]
        upstream_field_at_stations = np.moveaxis(upstream_field_at_stations, -1, 0)

        return dict(zip(points, upstream_field_at_stations))

    points = points_to_numpy(points)

    stations_1d = points_to_1d_indices(river_network, points)

    upstream_metric_field = calculate_upstream_metric(
        river_network,
        field,
        metric,
        weights,
        mv,
        accept_missing,
        skip=True,
    )

    metric_at_stations = upstream_metric_field[..., stations_1d]

    return {(x, y): metric_at_stations[..., i] for i, (x, y) in enumerate(zip(*points))}


@mask_and_unmask
def find(river_network, field, mv=0, in_place=False):
    """Labels the catchments given a field of labelled sinks.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    mv : scalar, optional
        The missing value indicator. Default is 0.
    in_place : bool, optional
        If True, modifies the input field in place. Default is False.

    Returns
    -------
    numpy.ndarray
        The field values accumulated downstream.

    """
    if not in_place:
        field = field.copy()

    if len(field.shape) == 1:
        op = _find_catchments_2D
    else:
        op = _find_catchments_ND

    def operation(river_network, field, grouping, mv):
        return op(river_network, field, grouping, mv, overwrite=True)

    return flow(river_network, field, True, operation, mv)


for metric in metrics_dict.keys():

    func = partial(calculate_catchment_metric, metric=metric)

    globals()[metric] = func


def _find_catchments_2D(river_network, field, grouping, mv, overwrite):
    """Updates field in-place with the value of its downstream nodes, dealing
    with missing values for 2D fields.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    grouping : numpy.ndarray
        The array of node indices.
    mv : scalar
        The missing value indicator.
    overwrite : bool
        If True, overwrite existing non-missing values in the field array.

    Returns
    -------
    None

    """
    valid_group = grouping[
        ~is_missing(field[..., river_network.downstream_nodes[grouping]], mv)
    ]  # only update nodes where the downstream belongs to a catchment
    if not overwrite:
        valid_group = valid_group[is_missing(field[..., valid_group], mv)]
    field[..., valid_group] = field[..., river_network.downstream_nodes[valid_group]]


def _find_catchments_ND(river_network, field, grouping, mv, overwrite):
    """Updates field in-place with the value of its downstream nodes, dealing
    with missing values for ND fields.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    grouping : numpy.ndarray
        The array of node indices.
    mv : scalar
        The missing value indicator.
    overwrite : bool
        If True, overwrite existing non-missing values in the field array.

    Returns
    -------
    None

    """
    field = field.T
    valid_mask = ~is_missing(field[river_network.downstream_nodes[grouping]], mv)
    valid_indices = np.array(np.where(valid_mask))
    valid_indices[0] = grouping[valid_indices[0]]
    if not overwrite:
        temp_valid_indices = valid_indices[0]
        valid_mask = is_missing(field[tuple(valid_indices)], mv)
        valid_indices = np.array(np.where(valid_mask))
        valid_indices[0] = temp_valid_indices[valid_indices[0]]
    downstream_valid_indices = valid_indices.copy()
    downstream_valid_indices[0] = river_network.downstream_nodes[
        downstream_valid_indices[0]
    ]
    field[tuple(valid_indices)] = field[tuple(downstream_valid_indices)]
    return field.T



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from functools import partial

import numpy as np

from .accumulation import calculate_online_metric
from .metrics import metrics_dict
from .utils import mask_and_unmask


@mask_and_unmask
def calculate_upstream_metric(
    river_network,
    field,
    metric,
    weights=None,
    mv=np.nan,
    accept_missing=False,
):
    """
    Calculates a metric for the field over all upstream values.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    metric : str
        Metric to compute. Options are "mean", "max", "min", "sum", "product"
    weights : ndarray, optional
        Used to weight the field when computing the metric. Default is None.
    mv : scalar, optional
        Missing value for the input field. Default is np.nan.
    accept_missing : bool, optional
        Whether or not to accept missing values in the input field. Default is False.

    Returns
    -------
    numpy.ndarray
        Output field.

    """

    return calculate_online_metric(
        river_network, field, metric, weights, mv, accept_missing, flow_direction="down"
    )


for metric in metrics_dict.keys():

    func = partial(calculate_upstream_metric, metric=metric)

    globals()[metric] = func



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np


def compute_topological_labels(
    sources: np.ndarray, sinks: np.ndarray, downstream_nodes: np.ndarray
):
    """Finds the topological distance labels for each node in
    downstream_nodes.

    Parameters
    ----------
    sources : numpy.ndarray
        The river network sources.
    sinks : numpy.ndarray
        The river network sinks.
    downstream_nodes : cnumpy.ndarray
        The river network downstream nodes.

    Returns
    -------
    numpy.ndarray
        Array of topological distance labels for each node.
    """
    n_nodes = downstream_nodes.shape[0]
    inlets = downstream_nodes[sources]
    labels = np.zeros(n_nodes, dtype=int)

    for n in range(1, n_nodes + 1):
        inlets = inlets[inlets != n_nodes]  # subset to valid nodes
        if inlets.shape[0] == 0:
            break
        labels[inlets] = n  # update furthest distance from source
        inlets = downstream_nodes[inlets]

    if inlets.shape[0] != 0:
        raise ValueError("River Network contains a cycle.")

    labels[sinks] = n - 1  # put all sinks in last group in topological ordering

    return labels



import numpy as np
import pytest
from test_inputs.accumulation import *
from test_inputs.readers import *
from utils import convert_to_2d

import earthkit.hydro as ekh


def get_numpy_function(function_name):
    return getattr(np, function_name)


@pytest.mark.parametrize(
    "river_network, input_field, flow_downstream, mv",
    [
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1a,
            upstream_metric_sum_1a,
            mv_1a,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1b,
            upstream_metric_sum_1b,
            mv_1b,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1c,
            upstream_metric_sum_1c,
            mv_1c,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1d,
            upstream_metric_sum_1d,
            mv_1d,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1e,
            upstream_metric_sum_1e,
            mv_1e,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1f,
            upstream_metric_sum_1f,
            mv_1f,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1g,
            upstream_metric_sum_1g,
            mv_1g,
        ),
        (
            ("cama_nextxy", cama_nextxy_2),
            input_field_2a,
            upstream_metric_sum_2a,
            mv_2a,
        ),
        (
            ("cama_nextxy", cama_nextxy_2),
            input_field_2b,
            upstream_metric_sum_2b,
            mv_2b,
        ),
        (
            ("cama_nextxy", cama_nextxy_2),
            input_field_2g,
            upstream_metric_sum_2g,
            mv_2g,
        ),
    ],
    indirect=["river_network"],
)
def test_upstream_metric_sum(river_network, input_field, flow_downstream, mv):
    output_field = ekh.calculate_upstream_metric(
        river_network, input_field, "sum", weights=None, mv=mv, accept_missing=True
    )
    print(output_field)
    print(flow_downstream)
    assert output_field.dtype == flow_downstream.dtype
    np.testing.assert_allclose(output_field, flow_downstream)
    np.testing.assert_allclose(
        output_field,
        ekh.flow_downstream(
            river_network,
            input_field,
            mv=mv,
            in_place=False,
            ufunc=np.add,
            accept_missing=True,
        ),
    )

    input_field = convert_to_2d(river_network, input_field, 0)
    flow_downstream = convert_to_2d(river_network, flow_downstream, 0)
    output_field = ekh.calculate_upstream_metric(
        river_network, input_field, "sum", weights=None, mv=mv, accept_missing=True
    )
    print(output_field)
    print(flow_downstream)
    assert output_field.dtype == flow_downstream.dtype
    np.testing.assert_allclose(output_field, flow_downstream)
    np.testing.assert_allclose(
        output_field,
        ekh.flow_downstream(
            river_network,
            input_field,
            mv=mv,
            in_place=False,
            ufunc=np.add,
            accept_missing=True,
        ),
    )


@pytest.mark.parametrize(
    "river_network, input_field, flow_downstream, mv",
    [
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1a,
            upstream_metric_max_1a,
            mv_1a,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1b,
            upstream_metric_max_1b,
            mv_1b,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1c,
            upstream_metric_max_1c,
            mv_1c,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1d,
            upstream_metric_max_1d,
            mv_1d,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1e,
            upstream_metric_max_1e,
            mv_1e,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1f,
            upstream_metric_max_1f,
            mv_1f,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1g,
            upstream_metric_max_1g,
            mv_1g,
        ),
        (
            ("cama_nextxy", cama_nextxy_2),
            input_field_2a,
            upstream_metric_max_2a,
            mv_2a,
        ),
        (
            ("cama_nextxy", cama_nextxy_2),
            input_field_2b,
            upstream_metric_max_2b,
            mv_2b,
        ),
    ],
    indirect=["river_network"],
)
def test_calculate_upstream_metric_max(river_network, input_field, flow_downstream, mv):
    output_field = ekh.calculate_upstream_metric(
        river_network, input_field, "max", weights=None, mv=mv, accept_missing=True
    )
    print(output_field)
    print(flow_downstream)
    assert output_field.dtype == flow_downstream.dtype
    np.testing.assert_allclose(output_field, flow_downstream)
    np.testing.assert_allclose(
        output_field,
        ekh.flow_downstream(
            river_network,
            input_field,
            mv,
            in_place=False,
            ufunc=np.maximum,
            accept_missing=True,
        ),
    )


@pytest.mark.parametrize(
    "river_network, input_field, flow_downstream, mv",
    [
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1a,
            upstream_metric_min_1a,
            mv_1a,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1b,
            upstream_metric_min_1b,
            mv_1b,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1c,
            upstream_metric_min_1c,
            mv_1c,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1d,
            upstream_metric_min_1d,
            mv_1d,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1e,
            upstream_metric_min_1e,
            mv_1e,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1f,
            upstream_metric_min_1f,
            mv_1f,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1g,
            upstream_metric_min_1g,
            mv_1g,
        ),
        (
            ("cama_nextxy", cama_nextxy_2),
            input_field_2a,
            upstream_metric_min_2a,
            mv_2a,
        ),
        (
            ("cama_nextxy", cama_nextxy_2),
            input_field_2b,
            upstream_metric_min_2b,
            mv_2b,
        ),
    ],
    indirect=["river_network"],
)
def test_calculate_upstream_metric_min(river_network, input_field, flow_downstream, mv):
    output_field = ekh.calculate_upstream_metric(
        river_network, input_field, "min", weights=None, mv=mv, accept_missing=True
    )
    print(output_field)
    print(flow_downstream)
    assert output_field.dtype == flow_downstream.dtype
    np.testing.assert_allclose(output_field, flow_downstream)
    np.testing.assert_allclose(
        output_field,
        ekh.flow_downstream(
            river_network,
            input_field,
            mv,
            in_place=False,
            ufunc=np.minimum,
            accept_missing=True,
        ),
    )


@pytest.mark.parametrize(
    "river_network, input_field, flow_downstream, mv",
    [
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1a,
            upstream_metric_product_1a,
            mv_1a,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1b,
            upstream_metric_product_1b,
            mv_1b,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1c,
            upstream_metric_product_1c,
            mv_1c,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1d,
            upstream_metric_product_1d,
            mv_1d,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1e,
            upstream_metric_product_1e,
            mv_1e,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1f,
            upstream_metric_product_1f,
            mv_1f,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1g,
            upstream_metric_product_1g,
            mv_1g,
        ),
        (
            ("cama_nextxy", cama_nextxy_2),
            input_field_2a,
            upstream_metric_product_2a,
            mv_2a,
        ),
        (
            ("cama_nextxy", cama_nextxy_2),
            input_field_2b,
            upstream_metric_product_2b,
            mv_2b,
        ),
    ],
    indirect=["river_network"],
)
def test_calculate_upstream_metric_prod(
    river_network, input_field, flow_downstream, mv
):
    output_field = ekh.calculate_upstream_metric(
        river_network, input_field, "prod", weights=None, mv=mv, accept_missing=True
    )
    print(output_field)
    print(flow_downstream)
    assert output_field.dtype == flow_downstream.dtype
    np.testing.assert_allclose(output_field, flow_downstream)
    np.testing.assert_allclose(
        output_field,
        ekh.flow_downstream(
            river_network,
            input_field,
            mv,
            in_place=False,
            ufunc=np.multiply,
            accept_missing=True,
        ),
    )


@pytest.mark.parametrize(
    "river_network, input_field, flow_downstream, mv",
    [
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1a,
            upstream_metric_mean_1a,
            mv_1a,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1b,
            upstream_metric_mean_1b,
            mv_1b,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1c,
            upstream_metric_mean_1c,
            mv_1c,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1d,
            upstream_metric_mean_1d,
            mv_1d,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1e,
            upstream_metric_mean_1e,
            mv_1e,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1f,
            upstream_metric_mean_1f,
            mv_1f,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            input_field_1g,
            upstream_metric_mean_1g,
            mv_1g,
        ),
        (
            ("cama_nextxy", cama_nextxy_2),
            input_field_2a,
            upstream_metric_mean_2a,
            mv_2a,
        ),
        (
            ("cama_nextxy", cama_nextxy_2),
            input_field_2b,
            upstream_metric_mean_2b,
            mv_2b,
        ),
    ],
    indirect=["river_network"],
)
def test_calculate_upstream_metric_mean(
    river_network, input_field, flow_downstream, mv
):
    output_field = ekh.calculate_upstream_metric(
        river_network, input_field, "mean", weights=None, mv=mv, accept_missing=True
    )
    assert output_field.dtype == flow_downstream.dtype
    np.testing.assert_allclose(output_field, flow_downstream)

    input_field = convert_to_2d(river_network, input_field, 0)
    flow_downstream = convert_to_2d(river_network, flow_downstream, 0)
    output_field = ekh.calculate_upstream_metric(
        river_network, input_field, "mean", weights=None, mv=mv, accept_missing=True
    )
    print(output_field)
    print(flow_downstream)
    assert output_field.dtype == flow_downstream.dtype
    np.testing.assert_allclose(output_field, flow_downstream)


@pytest.mark.parametrize(
    "river_network, input_field, accum_field",
    [
        (("d8_ldd", d8_ldd_1), input_field_1b, upstream_metric_sum_1g),
        (("cama_downxy", cama_downxy_1), input_field_1b, upstream_metric_sum_1g),
        (("cama_nextxy", cama_nextxy_1), input_field_1b, upstream_metric_sum_1g),
        (("d8_ldd", d8_ldd_2), input_field_2b, upstream_metric_sum_2g),
        (("cama_downxy", cama_downxy_2), input_field_2b, upstream_metric_sum_2g),
        (("cama_nextxy", cama_nextxy_2), input_field_2b, upstream_metric_sum_2g),
    ],
    indirect=["river_network"],
)
def test_accumulate_downstream_missing(river_network, input_field, accum_field):
    accum = ekh.flow_downstream(river_network, input_field, mv=-1, accept_missing=True)
    print(accum)
    print(accum_field)
    np.testing.assert_array_equal(accum, accum_field)


@pytest.mark.parametrize(
    "river_network",
    [
        ("d8_ldd", d8_ldd_1),
        ("cama_downxy", cama_downxy_1),
        ("cama_nextxy", cama_nextxy_1),
        ("d8_ldd", d8_ldd_2),
        ("cama_downxy", cama_downxy_2),
        ("cama_nextxy", cama_nextxy_2),
    ],
    indirect=True,
)
@pytest.mark.parametrize("N", range(4))
def test_accumulate_downstream_2d(river_network, N):
    field = np.random.rand(*([np.random.randint(10)] * N), *river_network.mask.shape)
    field_1d = field[..., river_network.mask]
    accum = ekh.flow_downstream(river_network, field_1d)
    np.testing.assert_array_equal(
        accum, ekh.flow_downstream(river_network, field)[..., river_network.mask]
    )
    np.testing.assert_array_equal(
        ekh.flow_downstream(river_network, field)[..., ~river_network.mask],
        field[..., ~river_network.mask],
    )


@pytest.mark.parametrize(
    "river_network",
    [
        ("d8_ldd", d8_ldd_1),
        ("cama_downxy", cama_downxy_1),
        ("cama_nextxy", cama_nextxy_1),
        ("d8_ldd", d8_ldd_2),
        ("cama_downxy", cama_downxy_2),
        ("cama_nextxy", cama_nextxy_2),
    ],
    indirect=True,
)
@pytest.mark.parametrize("metric", ["sum", "max", "min", "mean", "std", "var"])
def test_calculate_catchment(river_network, metric):
    field = np.random.rand(3, 4, river_network.n_nodes)
    weights = np.random.rand(3, 4, river_network.n_nodes)
    catchment_metric = ekh.calculate_catchment_metric(
        river_network, field, river_network.nodes, metric, weights
    )
    upstream_field = ekh.calculate_upstream_metric(
        river_network, field, metric, weights
    )
    print(catchment_metric[0].shape, upstream_field.shape)
    for i in catchment_metric.keys():
        assert catchment_metric[i].dtype == upstream_field.dtype
        np.testing.assert_allclose(catchment_metric[i], upstream_field[..., i])


@pytest.mark.parametrize(
    "river_network",
    [
        ("d8_ldd", d8_ldd_1),
        ("cama_downxy", cama_downxy_1),
        ("cama_nextxy", cama_nextxy_1),
        ("d8_ldd", d8_ldd_2),
        ("cama_downxy", cama_downxy_2),
        ("cama_nextxy", cama_nextxy_2),
    ],
    indirect=True,
)
@pytest.mark.parametrize("metric", ["sum", "max", "min", "mean", "prod", "std", "var"])
def test_calculate_subcatchment(river_network, metric):
    shape = (3, 4)
    field = np.random.rand(*shape, river_network.n_nodes)
    weights = np.random.rand(*shape, river_network.n_nodes)
    subcatchment_metric = ekh.calculate_subcatchment_metric(
        river_network, field, river_network.sinks, metric, weights
    )
    catchment_metric = ekh.calculate_catchment_metric(
        river_network, field, river_network.sinks, metric, weights
    )
    for i in catchment_metric.keys():
        assert catchment_metric[i].dtype == subcatchment_metric[i].dtype
        # the two methods don't match always because of numerical instabilities in the
        # variance computation arising from the different scales
        np.testing.assert_allclose(
            subcatchment_metric[i], catchment_metric[i], atol=5e-8
        )

    subcatchment_metric = ekh.calculate_subcatchment_metric(
        river_network, field, river_network.nodes, metric, weights
    )
    catchment_metric = ekh.calculate_catchment_metric(
        river_network,
        field,
        np.array(
            [
                i
                for i in river_network.nodes
                if i not in river_network.sinks and i not in river_network.sources
            ]
        ),
        metric,
        weights,
    )
    for i in catchment_metric.keys():
        assert catchment_metric[i].dtype == subcatchment_metric[i].dtype
        np.testing.assert_raises(
            AssertionError,
            np.testing.assert_allclose,
            subcatchment_metric[i],
            catchment_metric[i],
        )


@pytest.mark.parametrize(
    "river_network",
    [
        ("d8_ldd", d8_ldd_1),
        ("cama_downxy", cama_downxy_1),
        ("cama_nextxy", cama_nextxy_1),
        ("d8_ldd", d8_ldd_2),
        ("cama_downxy", cama_downxy_2),
        ("cama_nextxy", cama_nextxy_2),
    ],
    indirect=True,
)
@pytest.mark.parametrize("metric", ["sum", "max", "min", "mean", "prod", "std", "var"])
def test_calculate_catchment_numpy(river_network, metric):
    shape = (3, 4)
    field = np.random.rand(*shape, river_network.n_nodes)
    weights = None
    nodes = np.random.choice(river_network.n_nodes, 5)
    catchment_metric = ekh.calculate_catchment_metric(
        river_network, field, nodes, metric, weights
    )
    numpy_func = get_numpy_function(metric)
    for i in catchment_metric.keys():
        mask = np.zeros(river_network.n_nodes, dtype=bool)
        mask[i] = True
        mask = ekh.catchments.find(river_network, mask)
        numpy_answer = numpy_func(field[..., mask], axis=-1)

        np.testing.assert_allclose(catchment_metric[i], numpy_answer)


@pytest.mark.parametrize(
    "river_network",
    [
        ("d8_ldd", d8_ldd_1),
        ("cama_downxy", cama_downxy_1),
        ("cama_nextxy", cama_nextxy_1),
        ("d8_ldd", d8_ldd_2),
        ("cama_downxy", cama_downxy_2),
        ("cama_nextxy", cama_nextxy_2),
    ],
    indirect=True,
)
@pytest.mark.parametrize("metric", ["sum", "max", "min", "mean", "prod", "std", "var"])
def test_calculate_subcatchment_numpy(river_network, metric):
    shape = (3, 4)
    field = np.random.rand(*shape, river_network.n_nodes)
    weights = None
    nodes = np.random.choice(river_network.n_nodes, 5)
    subcatchment_metric = ekh.calculate_subcatchment_metric(
        river_network, field, nodes, metric, weights
    )
    numpy_func = get_numpy_function(metric)
    labels = np.zeros(river_network.n_nodes, dtype=int)
    labels[nodes] = nodes + 1
    labels = ekh.subcatchments.find(river_network, labels)
    for i in subcatchment_metric.keys():
        mask = labels == (i + 1)
        numpy_answer = numpy_func(field[..., mask], axis=-1)

        np.testing.assert_allclose(subcatchment_metric[i], numpy_answer)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from functools import partial

import numpy as np

from .catchments import _find_catchments_2D, _find_catchments_ND
from .core import flow
from .metrics import metrics_dict
from .utils import mask_2d, mask_and_unmask, points_to_1d_indices, points_to_numpy
from .zonal import calculate_zonal_metric


@mask_2d
def calculate_subcatchment_metric(
    river_network,
    field,
    points,
    metric,
    weights=None,
    mv=np.nan,
    accept_missing=False,
):
    """
    Calculates the metric over the subcatchments defined by stations.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    stations : list of tuples
        List of tuple indices of the stations.
    metric : str
        Metric to compute. Options are "mean", "max", "min", "sum", "prod",
        "std", "var".
    weights : ndarray
        Used to weight the field when computing the metric. Default is None.
    mv : scalar
        Missing value for the input field. Default is np.nan.
    accept_missing : bool
        Whether or not to accept missing values in the input field. Default is False.

    Returns
    -------
    dict
        Dictionary with (station, catchment_metric) pairs.
    """
    if isinstance(points, np.ndarray):
        initial_field = np.zeros(river_network.n_nodes, dtype=int)
        initial_field[points] = np.arange(points.shape[0]) + 1
        labels = find(river_network, initial_field, skip=True)
        metric_at_stations = calculate_zonal_metric(
            field,
            labels,
            metric,
            weights if weights is not None else None,
            mv,
            0,  # missing labels value
            accept_missing,
        )
        return {x: metric_at_stations[y] for (x, y) in zip(points, labels[points])}

    points = points_to_numpy(points)

    stations_1d = points_to_1d_indices(river_network, points)

    initial_field = np.zeros(river_network.n_nodes, dtype=int)
    unique_labels = np.arange(stations_1d.shape[0]) + 1
    initial_field[stations_1d] = unique_labels
    labels = find(river_network, initial_field, skip=True)
    metric_at_stations = calculate_zonal_metric(
        field,
        labels,
        metric,
        weights if weights is not None else None,
        mv,
        0,  # missing labels value
        accept_missing,
    )
    return {(x, y): metric_at_stations[z] for (x, y, z) in zip(*points, unique_labels)}


@mask_and_unmask
def find(river_network, field, mv=0, in_place=False):
    """Labels the catchments given a field of labelled sinks.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : numpy.ndarray
        The input field.
    mv : scalar, optional
        The missing value indicator. Default is 0.
    in_place : bool, optional
        If True, modifies the input field in place. Default is False.

    Returns
    -------
    numpy.ndarray
        The field values accumulated downstream.

    """
    if not in_place:
        field = field.copy()

    if len(field.shape) == 1:
        op = _find_catchments_2D
    else:
        op = _find_catchments_ND

    def operation(river_network, field, grouping, mv):
        return op(river_network, field, grouping, mv, overwrite=False)

    return flow(river_network, field, True, operation, mv)


for metric in metrics_dict.keys():

    func = partial(calculate_subcatchment_metric, metric=metric)

    globals()[metric] = func



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os
import tempfile
from hashlib import sha256

import joblib
import numpy as np

from ._version import __version__ as ekh_version
from .network_class import RiverNetwork

# read in only up to second decimal point
# i.e. 0.1.dev90+gfdf4e33.d20250107 -> 0.1
ekh_version = ".".join(ekh_version.split(".")[:2])


def cache(func):
    """Decorator to allow automatic use of cache.

    Parameters
    ----------
    func : callable
        The function to be wrapped and executed with masking applied.

    Returns
    -------
    callable
        The wrapped function.

    """

    def wrapper(
        path,
        river_network_format,
        source="file",
        use_cache=True,
        cache_dir=tempfile.mkdtemp(suffix="_earthkit_hydro"),
        cache_fname="{ekh_version}_{hash}.joblib",
        cache_compression=1,
    ):
        """Wrapper to load river network from cache if available, otherwise
        create and cache it.

        Parameters
        ----------
        path : str
            The path to the river network.
        river_network_format : str
            The format of the river network file.
            Supported formats are "precomputed", "cama", "pcr_d8", and "esri_d8".
        source : str, optional
            The source of the river network.
            For possible sources see:
            https://earthkit-data.readthedocs.io/en/latest/guide/sources.html
        use_cache : bool, optional
            Whether to use caching. Default is True.
        cache_dir : str, optional
            The directory to store the cache files. Default is a temporary directory.
        cache_fname : str, optional
            The filename template for the cache files.
            Default is "{ekh_version}_{hash}.joblib".
        cache_compression : int, optional
            The compression level for the cache files. Default is 1.

        Returns
        -------
        earthkit.hydro.RiverNetwork
            The loaded river network.

        """
        if use_cache:
            hashed_name = sha256(path.encode("utf-8")).hexdigest()
            cache_dir = cache_dir.format(ekh_version=ekh_version, hash=hashed_name)
            cache_fname = cache_fname.format(ekh_version=ekh_version, hash=hashed_name)
            cache_filepath = os.path.join(cache_dir, cache_fname)

            if os.path.isfile(cache_filepath):
                print(f"Loading river network from cache ({cache_filepath}).")
                return joblib.load(cache_filepath)
            else:
                print(f"River network not found in cache ({cache_filepath}).")
                os.makedirs(cache_dir, exist_ok=True)
        else:
            print("Cache disabled.")

        network = func(path, river_network_format, source)

        if use_cache:
            joblib.dump(network, cache_filepath, compress=cache_compression)
            print(f"River network loaded, saving to cache ({cache_filepath}).")

        return network

    return wrapper


def import_earthkit_or_prompt_install(river_network_format, source):
    """Ensure that the `earthkit.data` package is installed and import it. If
    the package is not installed, prompt the user to install it.

    Parameters
    ----------
    river_network_format : str
        The format of the river network file.
    source : str
        The source of the river network.

    Returns
    -------
    module
        The imported `earthkit.data` module.

    Raises
    ------
    ModuleNotFoundError
        If the `earthkit.data` package is not installed.

    """
    try:
        import earthkit.data as ekd
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "earthkit-data is required for loading river network format"
            f"{river_network_format} from source {source}."
            "\nTo install it, run `pip install earthkit-data`"
        )
    return ekd


def find_main_var(ds, min_dim=2):
    """Find the main variable in the dataset.

    Parameters
    ----------
    ds : xarray.Dataset
        The dataset to search for the main variable.
    min_dim : int, optional
        The minimum number of dimensions the variable must have. Default is 2.

    Returns
    -------
    str
        The name of the main variable.

    Raises
    ------
    ValueError
        If no variable or more than one variable with the required dimensions is found.

    """
    variable_names = [k for k in ds.variables if len(ds.variables[k].dims) >= min_dim]
    if len(variable_names) > 1:
        raise ValueError("More than one variable of dimension >= {min_dim} in dataset.")
    elif len(variable_names) == 0:
        raise ValueError("No variable of dimension >= {min_dim} in dataset.")
    else:
        return variable_names[0]


def from_cama_nextxy(x, y):
    """Create a river network from CaMa nextxy data.

    Parameters
    ----------
    x : numpy.ndarray
        The x-coordinates of the next downstream cell.
    y : numpy.ndarray
        The y-coordinates of the next downstream cell.

    Returns
    -------
    earthkit.hydro.RiverNetwork
        The created river network.

    """
    shape = x.shape
    x = x.flatten()
    missing_mask = x != -9999
    mask_upstream = ((x != -9) & (x != -9999)) & (x != -10)
    upstream_indices = np.arange(x.size)[mask_upstream]
    x = x[mask_upstream]
    y = y.flatten()[mask_upstream]
    x -= 1
    y -= 1
    downstream_indices = x + y * shape[1]
    return create_network(upstream_indices, downstream_indices, missing_mask, shape)


def from_cama_downxy(dx, dy):
    """Create a river network from CaMa downxy data.

    Parameters
    ----------
    dx : numpy.ndarray
        The x-offsets of the next downstream cell.
    dy : numpy.ndarray
        The y-offsets of the next downstream cell.

    Returns
    -------
    earthkit.hydro.RiverNetwork
        The created river network.

    """
    x_offsets = dx
    y_offsets = dy.flatten()
    shape = x_offsets.shape
    x_offsets = x_offsets.flatten()
    mask_upstream = ((x_offsets != -999) & (x_offsets != -9999)) & (x_offsets != -1000)
    missing_mask = x_offsets != -9999
    x_offsets = x_offsets[mask_upstream]
    y_offsets = y_offsets[mask_upstream]
    upstream_indices, downstream_indices = (
        find_upstream_downstream_indices_from_offsets(
            x_offsets, y_offsets, missing_mask, mask_upstream, shape
        )
    )
    return create_network(upstream_indices, downstream_indices, missing_mask, shape)


def from_d8(data, river_network_format="pcr_d8"):
    """Create a river network from PCRaster d8 data.

    Parameters
    ----------
    data : numpy.ndarray
        The PCRaster d8 drain direction data.

    Returns
    -------
    earthkit.hydro.RiverNetwork
        The created river network.

    """
    shape = data.shape
    data_flat = data.flatten()
    del data
    if river_network_format == "pcr_d8":
        missing_mask = np.isin(data_flat, range(1, 10))
        mask_upstream = data_flat != 5
    elif river_network_format == "esri_d8":
        missing_mask = np.isin(data_flat, np.append(0, 2 ** np.arange(8))) & (
            data_flat != 255
        )
        mask_upstream = (data_flat != 0) & (data_flat != -1)
    elif river_network_format == "merit_d8":
        missing_mask = np.isin(data_flat, np.append(0, 2 ** np.arange(8))) & (
            data_flat != 247
        )
        mask_upstream = (data_flat != 0) & (data_flat != 255)
    else:
        raise ValueError(f"Unsupported river network format: {river_network_format}.")
    mask_upstream = (mask_upstream) & (missing_mask)
    directions = data_flat[mask_upstream].astype("int")
    del data_flat
    if river_network_format == "pcr_d8":
        x_offsets = np.array([0, -1, 0, +1, -1, 0, +1, -1, 0, +1])[directions]
        y_offsets = -np.array([0, -1, -1, -1, 0, 0, 0, 1, 1, 1])[directions]
    elif river_network_format == "esri_d8" or river_network_format == "merit_d8":
        x_mapping = {32: -1, 64: 0, 128: +1, 16: -1, 1: +1, 8: -1, 4: 0, 2: +1}
        y_mapping = {32: 1, 64: 1, 128: 1, 16: 0, 1: 0, 8: -1, 4: -1, 2: -1}
        x_offsets = np.vectorize(x_mapping.get)(directions)
        y_offsets = -np.vectorize(y_mapping.get)(directions)
    del directions
    upstream_indices, downstream_indices = (
        find_upstream_downstream_indices_from_offsets(
            x_offsets, y_offsets, missing_mask, mask_upstream, shape
        )
    )
    return create_network(upstream_indices, downstream_indices, missing_mask, shape)


def find_upstream_downstream_indices_from_offsets(
    x_offsets, y_offsets, missing_mask, mask_upstream, shape
):
    """Function to convert from offsets to absolute indices.

    Parameters
    ----------
    x_offsets : numpy.ndarray
        The x-offsets of the next downstream cell.
    y_offsets : numpy.ndarray
        The y-offsets of the next downstream cell.
    missing_mask : numpy.ndarray
        A boolean mask indicating missing values in the data.
    mask_upstream : numpy.ndarray
        A boolean mask indicating upstream cells.
    shape : tuple
        The shape of the original data array.

    Returns
    -------
    earthkit.hydro.RiverNetwork
        The created river network.

    """
    ny, nx = shape
    upstream_indices = np.arange(missing_mask.size)[mask_upstream]
    del mask_upstream
    x_coords = upstream_indices % nx
    x_coords = (x_coords + x_offsets) % nx
    downstream_indices = x_coords
    del x_coords
    y_coords = np.floor_divide(upstream_indices, nx)
    y_coords = (y_coords + y_offsets) % ny
    downstream_indices += y_coords * nx
    del y_coords
    return upstream_indices, downstream_indices


def create_network(upstream_indices, downstream_indices, missing_mask, shape):
    """Creates a river network from upstream and downstream indices.

    Parameters
    ----------
    upstream_indices : numpy.ndarray
        Indices of upstream nodes.
    downstream_indices : numpy.ndarray
        Indices of downstream nodes.
    missing_mask : numpy.ndarray
        Boolean mask indicating the presence of nodes.
    shape : tuple
        Shape of the original domain.

    Returns
    -------
    earthkit.hydro.RiverNetwork
        The created river network.

    """
    n_nodes = int(np.sum(missing_mask))
    nodes = np.arange(n_nodes, dtype=np.uintp)
    nodes_matrix = np.ones(missing_mask.size, dtype=np.uintp) * n_nodes
    nodes_matrix[missing_mask] = nodes
    upstream_nodes = nodes_matrix[upstream_indices]
    downstream_nodes = nodes_matrix[downstream_indices]
    del upstream_indices, downstream_indices, nodes_matrix
    downstream = np.ones(n_nodes, dtype=np.uintp) * n_nodes
    downstream[upstream_nodes] = downstream_nodes
    del downstream_nodes, upstream_nodes, n_nodes
    return RiverNetwork(nodes, downstream, missing_mask.reshape(shape))



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np


class SumBased:
    func = np.add
    base_val = 0


class MaxBased:
    func = np.maximum
    base_val = -np.inf


class MinBased:
    func = np.minimum
    base_val = np.inf


class ProductBased:
    func = np.multiply
    base_val = 1


metrics_dict = {
    "sum": SumBased,
    "mean": SumBased,
    "max": MaxBased,
    "min": MinBased,
    "prod": ProductBased,
    "std": SumBased,
    "var": SumBased,
}



import numpy as np
import pytest
from test_inputs.distance import *
from test_inputs.readers import *

import earthkit.hydro as ekh


@pytest.mark.parametrize(
    "river_network, stations_list, upstream, downstream, weights, result",
    [
        (
            ("cama_nextxy", cama_nextxy_1),
            stations,
            True,
            True,
            weights_1,
            distance_1_min_up_down,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            stations,
            True,
            False,
            weights_1,
            distance_1_min_up,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            stations,
            False,
            True,
            weights_1,
            distance_1_min_down,
        ),
    ],
    indirect=["river_network"],
)
def test_distance_min(
    river_network, stations_list, upstream, downstream, weights, result
):
    dist = ekh.distance.min(
        river_network,
        stations_list,
        upstream=upstream,
        downstream=downstream,
        weights=weights,
    )
    np.testing.assert_allclose(dist, result)


@pytest.mark.parametrize(
    "river_network, stations_list, upstream, downstream, weights, result",
    [
        (
            ("cama_nextxy", cama_nextxy_1),
            stations,
            True,
            False,
            weights_1,
            distance_1_max_up,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            stations,
            False,
            True,
            weights_1,
            distance_1_max_down,
        ),
    ],
    indirect=["river_network"],
)
def test_distance_max(
    river_network, stations_list, upstream, downstream, weights, result
):
    dist = ekh.distance.max(
        river_network,
        stations_list,
        upstream=upstream,
        downstream=downstream,
        weights=weights,
    )
    np.testing.assert_allclose(dist, result)


@pytest.mark.parametrize(
    "river_network, stations_list, upstream, downstream, weights, result",
    [
        (
            ("cama_nextxy", cama_nextxy_1),
            stations,
            True,
            True,
            weights_1,
            length_1_min_up_down,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            stations,
            True,
            False,
            weights_1,
            length_1_min_up,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            stations,
            False,
            True,
            weights_1,
            length_1_min_down,
        ),
    ],
    indirect=["river_network"],
)
def test_length_min(
    river_network, stations_list, upstream, downstream, weights, result
):
    length = ekh.length.min(
        river_network,
        stations_list,
        upstream=upstream,
        downstream=downstream,
        weights=weights,
    )
    np.testing.assert_allclose(length, result)


@pytest.mark.parametrize(
    "river_network, stations_list, upstream, downstream, weights, result",
    [
        (
            ("cama_nextxy", cama_nextxy_1),
            stations,
            True,
            False,
            weights_1,
            length_1_max_up,
        ),
        (
            ("cama_nextxy", cama_nextxy_1),
            stations,
            False,
            True,
            weights_1,
            length_1_max_down,
        ),
    ],
    indirect=["river_network"],
)
def test_length_max(
    river_network, stations_list, upstream, downstream, weights, result
):
    length = ekh.length.max(
        river_network,
        stations_list,
        upstream=upstream,
        downstream=downstream,
        weights=weights,
    )
    np.testing.assert_allclose(length, result)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np


def is_missing(field, mv):
    """Finds a mask of missing values.

    Parameters
    ----------
    field : numpy.ndarray
        The scalar input field to check for missing values.
    mv : scalar
        The missing value to check for.

    Returns
    -------
    numpy.ndarray
        A boolean mask of missing values.

    """
    if np.isnan(mv):
        return np.isnan(field)
    elif np.isinf(mv):
        return np.isinf(field)
    else:
        return field == mv


def are_missing_values_present(field, mv):
    """Finds if missing values are present in a field.

    Parameters
    ----------
    field : numpy.ndarray
        The scalar input field to check for missing values.
    mv : scalar
        The missing value to check for.

    Returns
    -------
    bool
        True if missing values are present, False otherwise.

    """
    return np.any(is_missing(field, mv))


def check_missing(field, mv, accept_missing):
    """Finds missing values and checks if they are allowed in the input field.

    Parameters
    ----------
    field : numpy.ndarray
        The scalar input field to check for missing values.
    mv : scalar
        The missing value to check for.
    accept_missing : bool
        If True, missing values are allowed in the input field.

    Returns
    -------
    bool
        True if missing values are present, False otherwise.

    """
    missing_values_present = are_missing_values_present(field, mv)
    if missing_values_present:
        if not accept_missing:
            raise ValueError(
                "Missing values present in input field and accept_missing is False."
            )
        else:
            print("Warning: missing values present in input field.")
    return missing_values_present


def missing_to_nan(field, mv, accept_missing, skip=False):
    """
    Converts a field with arbitrary missing values to a field of type float with nans.

    Parameters
    ----------
    field : numpy.ndarray
       The input field.
    mv : scalar
        Missing values for the input field.
    accept_missing : bool
        If True, missing values are allowed in the input field.
    skip : bool, optional
        Skip this function. Default is False.

    Returns
    -------
    numpy.ndarray
        Output field.
    numpy.dtype
        dtype of the original field.


    """
    if skip:
        return field, field.dtype

    missing_mask = is_missing(field, mv)

    if np.any(missing_mask):
        if not accept_missing:
            raise ValueError(
                "Missing values present in input field and accept_missing is False."
            )
        else:
            print("Warning: missing values present in input field.")

    field_dtype = field.dtype
    if not field_dtype == np.float64:
        field = field.astype(np.float64, copy=False)  # convert to float64
    if np.isnan(mv):
        return field, field_dtype
    field[missing_mask] = np.nan
    return field, field_dtype


def nan_to_missing(out_field, field_dtype, mv):
    """
    Converts a floating field with np.nans back to original field
    with original missing values.

    Parameters
    ----------
    out_field : numpy.ndarray
       Field of type float with np.nans.
    field_dtype : numpy.dtype
        dtype to convert to.
    mv : scalar
        Original missing values.

    Returns
    -------
    numpy.ndarray
        Output field.

    """
    if not np.isnan(mv):
        np.nan_to_num(out_field, copy=False, nan=mv)
    if field_dtype != np.float64:
        out_field = out_field.astype(field_dtype, copy=False)
    return out_field


def mask_2d(func):
    """Decorator to allow function to mask 2d inputs to the river network.

    Parameters
    ----------
    func : callable
        The function to be wrapped and executed with masking applied.

    Returns
    -------
    callable
        The wrapped function.

    """

    def wrapper(river_network, *args, **kwargs):
        """Wrapper masking 2d data fields to allow for processing along the
        river network, then undoing the masking.

        Parameters
        ----------
        river_network : object
            The RiverNetwork instance calling the method.
        *args : tuple
            Positional arguments passed to the wrapped function.
        **kwargs : dict
            Keyword arguments passed to the wrapped function.

        Returns
        -------
        numpy.ndarray
            The processed field.

        """
        args = tuple(
            (
                arg[..., river_network.mask]
                if isinstance(arg, np.ndarray)
                and arg.shape[-2:] == river_network.mask.shape
                else arg if isinstance(arg, np.ndarray) else arg
            )
            for arg in args
        )

        kwargs = {
            key: (
                value[..., river_network.mask]
                if isinstance(value, np.ndarray)
                and value.shape[-2:] == river_network.mask.shape
                else value if isinstance(value, np.ndarray) else value
            )
            for key, value in kwargs.items()
        }

        return func(river_network, *args, **kwargs)

    return wrapper


def mask_and_unmask(func):
    """Decorator to convert masked 2d inputs back to 1d.

    Parameters
    ----------
    func : callable
        The function to be wrapped and executed with masking applied.

    Returns
    -------
    callable
        The wrapped function.

    """

    def wrapper(river_network, field, *args, **kwargs):
        """Wrapper masking 2d data fields to allow for processing along the
        river network, then undoing the masking.

        Parameters
        ----------
        river_network : object
            The RiverNetwork instance calling the method.
        field : numpy.ndarray
            The input data field to be processed.
        *args : tuple
            Positional arguments passed to the wrapped function.
        **kwargs : dict
            Keyword arguments passed to the wrapped function.

        Returns
        -------
        numpy.ndarray
            The processed field.

        """

        # skip! don't bother masking and unmasking
        # (if it has already been done)
        skip = kwargs.pop("skip", None)
        skip = skip if skip is not None else False
        if skip:
            return func(river_network, field, *args, **kwargs)

        # gets the missing value from the keyword arguments if it is present,
        # otherwise takes default value of mv from func
        mv = kwargs.get("mv")
        mv = mv if mv is not None else func.__defaults__[0]
        if field.shape[-2:] == river_network.mask.shape:
            in_place = kwargs.get("in_place", False)

            values_on_river_network = mask_2d(func)(
                river_network, field, *args, **kwargs
            )

            if in_place:
                out_field = field
            else:
                out_field = np.empty(field.shape, dtype=values_on_river_network.dtype)

            out_field[..., river_network.mask] = values_on_river_network

            if np.result_type(mv, field) != field.dtype:
                raise ValueError(
                    f"Missing value of type {type(mv)} is not compatible"
                    f" with field of dtype {field.dtype}"
                )

            out_field[..., ~river_network.mask] = mv
            return out_field
        else:
            return mask_2d(func)(river_network, field, *args, **kwargs)

    return wrapper


def points_to_numpy(points):
    """
    Converts a list of tuples (indices) into a tuple of lists
    for use in numpy indexing.

    Parameters
    ----------
    points : list
        List of tuple indices of the points.

    Returns
    -------
    tuple
        Tuple of points suitable for numpy indexing.
    """
    # transform here list of tuples (indices) into a tuple of lists
    # (easier to manipulate)
    points = np.array(points)
    return (points[:, 0], points[:, 1])


def points_to_1d_indices(river_network, stations):
    """ "
    Converts a numpy index into a 1D index suitable
    for use with the flattened river representation.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        The RiverNetwork instance calling the method.
    stations : tuple
        Tuple of numpy arrays defining the points.

    Returns
    -------
    numpy.ndarray
        1D array of indices.
    """
    node_numbers = np.cumsum(river_network.mask) - 1
    valid_stations = river_network.mask[stations]
    if np.any(~valid_stations):
        raise ValueError("Not all points are present on the river network.")
    stations = tuple(station_index[valid_stations] for station_index in stations)
    stations_1d = node_numbers.reshape(river_network.mask.shape)[stations]
    return stations_1d



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


def flow(river_network, field, invert_graph, operation, *args, **kwargs):
    """Apply an operation to a field along a river network.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    field : ndarray
        The field data to be modified in place.
    invert_graph : bool
        If True, process the river network from sinks to sources.
        If False, process from sources to sinks.
    operation : callable
        The operation to apply to the field. This function should
        take four arguments: river_network, field, grouping, and mv.
    *args, **kwargs : additional arguments
        Additional arguments to pass to the operation function.

    Returns
    -------
    ndarray
        The modified field after applying the operation along the river network.

    """
    if invert_graph:
        groupings = river_network.topological_groups[:-1][
            ::-1
        ]  # go from sinks to sources
    else:
        groupings = river_network.topological_groups[:-1]  # go from sources to sinks

    for grouping in groupings:
        # modify field in_place with desired operation
        # NB: this function needs to handle missing values
        # mv if they are allowed in input
        operation(river_network, field, grouping, *args, **kwargs)

    return field



import numpy as np

mask_2 = np.array(
    [
        True,
        True,
        True,
        True,
        True,
        True,
        False,
        True,
        True,
        True,
        True,
        True,
        True,
        False,
        True,
        True,
    ]
)


masked_unit_accuflux_2 = np.array([2, 1, 2, 1, 1, 2, 3, 1, 1, 3, 6, 1, 1, 2])



import numpy as np

downstream_nodes_1 = np.array(
    [
        5,
        6,
        7,
        7,
        8,
        10,
        11,
        12,
        12,
        13,
        16,
        16,
        16,
        12,
        13,
        16,
        20,  # we set sink to len of nodes
        16,
        17,
        18,
    ]
)


downstream_nodes_2 = np.array(
    [
        16,  # we set sink to len of nodes
        2,
        7,
        16,  # we set sink to len of nodes
        5,
        10,
        10,
        11,
        13,
        13,
        13,
        6,
        0,
        16,  # we set sink to len of nodes
        15,
        11,
    ]
)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np

from .accumulation import flow_downstream, flow_upstream
from .utils import mask_2d, points_to_1d_indices, points_to_numpy


@mask_2d
def min(
    river_network, points, weights=None, upstream=False, downstream=True, mv=np.nan
):
    """
    Calculate the minimum distance to a set of points in a river network.
    The distance is calculated along the river network, and can be
    computed in both/either upstream and downstream directions.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    points : list of tuples
        List of tuple indices of the points.
    weights : numpy.ndarray, optional
        Distance to the downstream point. Default is None, which
        corresponds to a unit distance for all points.
    upstream : bool, optional
        If True, calculates the distance in the upstream direction.
        Default is False.
    downstream : bool, optional
        If True, calculate the distance in the downstream direction.
        Default is True.
    mv : scalar, optional
        The missing value indicator. Default is np.nan.

    Returns
    -------
    numpy.ndarray
        The distance to the points in the river network.
    """

    if weights is None:
        weights = np.ones(river_network.n_nodes)
    else:
        # maybe check sinks are all zero or nan distance
        pass
    field = np.empty(river_network.n_nodes)
    field.fill(np.inf)

    if isinstance(points, np.ndarray):
        points_1d = points
    else:
        points = points_to_numpy(points)
        points_1d = points_to_1d_indices(river_network, points)

    field[points_1d] = 0
    if downstream:
        field = flow_downstream(
            river_network,
            field,
            mv,
            ufunc=np.minimum,
            additive_weight=weights,
            modifier_use_upstream=True,
        )
    if upstream:
        field = flow_upstream(
            river_network,
            field,
            mv,
            ufunc=np.minimum,
            additive_weight=weights,
            modifier_use_upstream=True,
        )

    out_field = np.empty(river_network.shape, dtype=field.dtype)
    out_field[..., river_network.mask] = field
    out_field[..., ~river_network.mask] = mv

    return out_field


@mask_2d
def max(
    river_network, points, weights=None, upstream=False, downstream=True, mv=np.nan
):
    """
    Calculate the maximum distance to a set of points in a river network.
    The distance is calculated along the river network, and can be
    computed in both/either upstream and downstream directions.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    points : list of tuples
        List of tuple indices of the points.
    weights : numpy.ndarray, optional
        Distance to the downstream point. Default is None, which
        corresponds to a unit distance for all points.
    upstream : bool, optional
        If True, calculates the distance in the upstream direction.
        Default is False.
    downstream : bool, optional
        If True, calculate the distance in the downstream direction.
        Default is True.
    mv : scalar, optional
        The missing value indicator. Default is np.nan.

    Returns
    -------
    numpy.ndarray
        The distance to the points in the river network.
    """

    if upstream and downstream:
        # TODO: define how this should work
        # can one overwrite a starting station's distance?
        #
        # NB: there is no nice way to do this as downstream
        # and then upstream like for min
        # because we would need to know the paths
        # to avoid looping over each other
        # Only way I think can think is doing each station
        # separately, but this will be very slow...
        raise NotImplementedError(
            "Max distance both upstream and downstream is not yet implemented."
        )

    weights = np.ones(river_network.n_nodes) if weights is None else weights

    field = np.empty(river_network.n_nodes)
    field.fill(-np.inf)

    if isinstance(points, np.ndarray):
        points_1d = points
    else:
        points = points_to_numpy(points)
        points_1d = points_to_1d_indices(river_network, points)

    field[points_1d] = 0

    if downstream:
        field = flow_downstream(
            river_network,
            field,
            mv,
            ufunc=np.maximum,
            additive_weight=weights,
            modifier_use_upstream=True,
        )
    if upstream:
        field = flow_upstream(
            river_network,
            field,
            mv,
            ufunc=np.maximum,
            additive_weight=weights,
            modifier_use_upstream=True,
        )

    field = np.nan_to_num(field, neginf=np.inf)

    out_field = np.empty(river_network.shape, dtype=field.dtype)
    out_field[..., river_network.mask] = field
    out_field[..., ~river_network.mask] = mv

    return out_field


def to_sink(river_network, weights=None, path="shortest", mv=np.nan):
    """
    Calculate the minimum or maximum distance to the sinks of a river network.
    The distance is calculated along the river network.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    weights : numpy.ndarray, optional
        Distance to the downstream point. Default is None, which
        corresponds to a unit distance for all points.
    path : str, optional
        Whether to find the distance of the shortest or longest path.
        Default is 'shortest'.
    mv : scalar, optional
        The missing value indicator. Default is np.nan.

    Returns
    -------
    numpy.ndarray
        The distance to the points in the river network.
    """

    if path == "shortest":
        return min(
            river_network,
            river_network.sinks,
            weights,
            upstream=True,
            downstream=False,
            mv=mv,
        )
    elif path == "longest":
        return max(
            river_network,
            river_network.sinks,
            weights,
            upstream=True,
            downstream=False,
            mv=mv,
        )
    else:
        raise ValueError("Path must be one of 'shortest' or 'longest'.")


def to_source(river_network, weights=None, path="shortest", mv=np.nan):
    """
    Calculate the minimum or maximum distance to the sources of a river network.
    The distance is calculated along the river network.

    Parameters
    ----------
    river_network : earthkit.hydro.RiverNetwork
        An earthkit-hydro river network object.
    weights : numpy.ndarray, optional
        Distance to the downstream point. Default is None, which
        corresponds to a unit distance for all points.
    path : str, optional
        Whether to find the distance of the shortest or longest path.
        Default is 'shortest'.
    mv : scalar, optional
        The missing value indicator. Default is np.nan.

    Returns
    -------
    numpy.ndarray
        The distance to the points in the river network.
    """

    if path == "shortest":
        return min(
            river_network,
            river_network.sources,
            weights,
            upstream=False,
            downstream=True,
            mv=mv,
        )
    elif path == "longest":
        return max(
            river_network,
            river_network.sources,
            weights,
            upstream=False,
            downstream=True,
            mv=mv,
        )
    else:
        raise ValueError("Path must be one of 'shortest' or 'longest'.")



import os

from setuptools import setup
from setuptools_rust import RustExtension

use_rust = int(os.environ.get("USE_RUST", "-1"))

if use_rust == 0:  # pure python
    # print("Building pure Python version.")
    rust_extensions = []
elif use_rust == 1:  # rust extension
    # print("Building with rust bindings.")
    rust_extensions = [RustExtension("earthkit.hydro._rust", "Cargo.toml")]
else:  # (default) try rust extension, if fail fallback to python
    # print("Building with rust bindings, and if failing reverting to pure Python.")
    rust_extensions = [
        RustExtension("earthkit.hydro._rust", "Cargo.toml", optional=True)
    ]

setup(rust_extensions=rust_extensions)



import numpy as np
import pytest
from test_inputs.movement import *
from test_inputs.readers import *

import earthkit.hydro as ekh


@pytest.mark.parametrize(
    "river_network, upstream",
    [
        (("d8_ldd", d8_ldd_1), upstream_1),
        (("cama_downxy", cama_downxy_1), upstream_1),
        (("cama_nextxy", cama_nextxy_1), upstream_1),
        (("d8_ldd", d8_ldd_2), upstream_2),
        (("cama_downxy", cama_downxy_2), upstream_2),
        (("cama_nextxy", cama_nextxy_2), upstream_2),
    ],
    indirect=["river_network"],
)
def test_upstream(river_network, upstream):
    field = np.arange(1, river_network.n_nodes + 1)
    ups = ekh.move_downstream(river_network, field)
    np.testing.assert_array_equal(ups, upstream)


@pytest.mark.parametrize(
    "river_network, downstream",
    [
        (("d8_ldd", d8_ldd_1), downstream_1),
        (("cama_downxy", cama_downxy_1), downstream_1),
        (("cama_nextxy", cama_nextxy_1), downstream_1),
        (("d8_ldd", d8_ldd_2), downstream_2),
        (("cama_downxy", cama_downxy_2), downstream_2),
        (("cama_nextxy", cama_nextxy_2), downstream_2),
    ],
    indirect=["river_network"],
)
def test_downstream(river_network, downstream):
    field = np.arange(1, river_network.n_nodes + 1)
    down = ekh.move_upstream(river_network, field)
    print(down)
    print(downstream)
    np.testing.assert_array_equal(down, downstream)


@pytest.mark.parametrize(
    "river_network, upstream",
    [
        (("d8_ldd", d8_ldd_1), upstream_1),
        (("cama_downxy", cama_downxy_1), upstream_1),
        (("cama_nextxy", cama_nextxy_1), upstream_1),
        (("d8_ldd", d8_ldd_2), upstream_2),
        (("cama_downxy", cama_downxy_2), upstream_2),
        (("cama_nextxy", cama_nextxy_2), upstream_2),
    ],
    indirect=["river_network"],
)
def test_upstream_ND(river_network, upstream):
    field = np.arange(1, river_network.n_nodes + 1)
    field = np.stack([field, field], axis=0)
    ups = ekh.move_downstream(river_network, field)
    np.testing.assert_array_equal(ups, np.stack([upstream, upstream], axis=0))


@pytest.mark.parametrize(
    "river_network, downstream",
    [
        (("d8_ldd", d8_ldd_1), downstream_1),
        (("cama_downxy", cama_downxy_1), downstream_1),
        (("cama_nextxy", cama_nextxy_1), downstream_1),
        (("d8_ldd", d8_ldd_2), downstream_2),
        (("cama_downxy", cama_downxy_2), downstream_2),
        (("cama_nextxy", cama_nextxy_2), downstream_2),
    ],
    indirect=["river_network"],
)
def test_downstream_ND(river_network, downstream):
    field = np.arange(1, river_network.n_nodes + 1)
    field = np.stack([field, field], axis=0)
    print(field.shape)
    ups = ekh.move_upstream(river_network, field)
    np.testing.assert_array_equal(ups, np.stack([downstream, downstream], axis=0))
